body {
  background-color: #EFECCA;
  color: #002F2F;
}
h1 {
  text-align: center
}
h1, h2, h3, h4, strong, th {
  color: #046380;
}
h2 {
  border-bottom: 1px solid #A7A37E;
}
table {
  border: 5px solid #b7b7b7;
  margin-bottom: 50px;
  background-color: #E6E2AF;
}
table.sortable thead:hover {
  border: 5px solid #A7A37E;
  margin-bottom: 50px;
  background-color: #E6E2AF;
}
td, th {
  margin-left: 0px;
  margin-right: 0px;
  padding-left: 8px;
  padding-right: 8px;
  padding-bottom: 2px;
  padding-top: 2px;
  text-align:left;
}
td {
  border-top: 5px solid #EFECCA;
  font-family: monospace;
  text-align: right;
}
td.net {
    background-color: #82C7DB;
}
td.offsets {
    background-color: green;
}
td.impacts {
    background-color: red;
}
tr:hover {
    background-color: #F4F0BB;
}
img {
  margin: 20px;
}
div.warning {
    text-align: center;
    color: #900;
    background-color: #fee;
    border: 3px solid #c00;
    padding: 20px;
    margin-bottom: 20px;
}
div.notice {
    text-align: center;
    color: #090;
    background-color: #efe;
    border: 3px solid #0c0;
    padding: 20px;
    margin-bottom: 20px;
}
td.negative{
    background-color: #F99;
    color: #900;
}
tr.hover td.negative {
    background-color: #FCC;
    color: #900;
}
td.positive{
    background-color: #9F9;
    color: #090;
}
tr.hover td.positive {
    background-color: #CFC;
    color: #090;
}

.center {
    text-align: center;
}

#toc {
    top: 0px;
    left: 0px;
    height: 100%;
    position: fixed;
    background: #333;
    box-shadow: inset -5px 0 5px 0px #000;
    width: 200px;
    padding-top: 20px;
    color: #fff;
    font-family: 'Voces',sans-serif;
}

#toc ul {
    margin: 0;
    padding: 0;
    list-style: none;
}

#toc li {
    padding: 5px 10px;
}

#toc a {
    color: #fff;
    text-decoration: none;
    display: block;
}

#toc .h2 {
    padding-left: 20px;
}

#toc .h3 {
    padding-left: 40px;
}

#toc .toc-active {
    background: #336699;
    box-shadow: inset -5px 0px 10px -5px #000;
}

#content {
    margin-left: 220px;
}
a.export, a.export:visited {
    text-decoration: none;
    color:#000;
    background-color:#ddd;
    border: 1px solid #ccc;
    padding:8px;
}

div#metadata {
    padding: 20px;
    text-align: right;
    display: inline-block;
    float: right;
    background-color: #efefef;
    border: 2px solid #cfcfcf;
    font-family: monospace;
    color: #696969;
}
